var async = require('async');
var fs = require('fs-extra');
var path = require('path');

var $ = require('cheerio');
var _ = require('lodash');

var VERSION = require('../../package.json').version;
var docsDir = path.join(__dirname, '../../docs');
var pageTitle = 'Methods:';

var docFilename = 'docs.html';
var mainModuleFile = 'module-drachtio-srf.html';
var mainScrollableSection = '#main-container';
var sectionTitleClass = '.page-title';

var HTMLFileBegin = '<!DOCTYPE html>\n<html lang="en">\n<head>\n';
var HTMLFileHeadBodyJoin = '</head>\n<body>';
var HTMLFileEnd = '</body>';

var additionalFooterText = ' Documentation has been modified from the original. ' +
    ' For more information, please see the <a href="https://github.com/caolan/async">async</a> repository.';

var tocStructure = require('../toc');

console.log(`tocStructure: ${JSON.stringify(tocStructure)}`);

function generateHTMLFile(filename, $page, callback) {
  var methodName = filename.match(/\/(\w+)\.js\.html$/);
  if (methodName) {
    var $thisMethodDocLink = $page.find('#toc').find('a[href="' + docFilename + '#' + methodName[1] + '"]');
    $thisMethodDocLink.parent().addClass('active');
  }

  // generate an HTML file from a cheerio object
  var HTMLdata = HTMLFileBegin + $page.find('head').html()
    + HTMLFileHeadBodyJoin + $page.find('body').html()
    + HTMLFileEnd;

  fs.writeFile(filename, HTMLdata, callback);
}

function extractModuleFiles(files) {
  console.log(`extractModuleFiles: input: ${files}`);
  const modules = _.filter(files, function(file) {
    return _.startsWith(file, 'module') && file !== mainModuleFile;
  });
  console.log(`extractModuleFiles: output: ${modules}`);
  return modules;
}


function produceClassArticle(className, $classPage, methods) {
  const $article = $('section article', $classPage);
  let preamble = true;
  let postamble = false;

  const $root = $(`<section id="${className.toLowerCase()}"></section>`);

  $root.append($('section header', $classPage));

  //console.log(`root: ${$root}`);

  const $end = $('<div class="postamble"></div>');
  let $section;
  const sections = {};
  let name;

  $article.children().each((i, elem) => {
    const $e = $(elem);
    const className = $e.attr('class');

    if (postamble) {
      $end.append($e);
    }
    else if (className === 'subsection-title' && $e.text().toLowerCase() === 'methods') {
      console.log('end of preamble');
      preamble = false;
      $root.append($e);
      $section = $('<div class="method-container"></div>');
    }
    else if (className === 'name' && !preamble && !postamble) {
      if (name) {
        console.log(`finished parsing ${name}`);
        $section.attr('id', name);
        sections[name] = $section;
      }
      name = $e.attr('id');
      $section = $('<div class="method-container"></div>');
      $section.append($e);
      console.log(`parsing method ${name}`);
    }
    else if (className === 'subsection-title' && !preamble) {
      console.log(`starting postamble with subsection-title ${$e.text()}`);
      postamble = true;
      $section.attr('id', name);
      sections[name] = $section;
      $end.append($e);
    }
    if (preamble) {
      $root.append($e);
    }
    else {
      $section.append($e);
    }
  });

  // add the methods in the order specified in toc.json
  _.each(methods, (m) => {
    console.log(`adding section ${m}`);
    $root.append(sections[m].children());
  });
  $root.append($end.find('.postamble').children());

  //console.log(`populated root: ${$root}`);
  return $root;

}

function combineClassDocs(files, callback) {
  //var moduleFiles = extractModuleFiles(files);

  const classes = [] ;
  _.each(tocStructure.classes, (val, key) => { classes.push(key);});
  console.log(`adding classes in this order ${classes}`);

  const mainModuleData = fs.readFileSync(path.join(docsDir, mainModuleFile), 'utf8');
  const $mainPage = $(mainModuleData);

  const data = {};

  classes.forEach((c) => {
    const filename = `${c}.html`;
    const classData = fs.readFileSync(path.join(docsDir, filename), 'utf8');
    const $classPage = $(classData);
    const className = $classPage.find(sectionTitleClass).text();
    $classPage.find(sectionTitleClass).attr('id', className.toLowerCase());
    //$mainPage.find(mainScrollableSection).append($classPage.find(mainScrollableSection).html());
    //console.log(`added content of ${filename} to ${mainModuleFile}`);

    const obj = produceClassArticle(c, $classPage, tocStructure.classes[c]);
    data[c] = obj;
  });

  console.log(`keys: ${_.keys(data)}`);
  _.each(data, (val, key) => {
    $mainPage.find(mainScrollableSection).append(val);
  });

  generateHTMLFile(path.join(docsDir, docFilename), $mainPage, callback);

    // each 'module' (category) has a separate page, with all of the
    // important information in a 'main' div. Combine all of these divs into
    // one on the actual module page
    /*
    async.eachSeries(moduleFiles, function(file, fileCallback) {
      fs.readFile(path.join(docsDir, file), 'utf8', function(err, moduleData) {
        if (err) return fileCallback(err);
        var $modulePage = $(moduleData);
        var moduleName = $modulePage.find(sectionTitleClass).text();
        $modulePage.find(sectionTitleClass).attr('id', moduleName.toLowerCase());
        $mainPage.find(mainScrollableSection).append($modulePage.find(mainScrollableSection).html());
        return fileCallback();
      });
    }, function(err) {
      if (err) return callback(err);
      generateHTMLFile(path.join(docsDir, docFilename), $mainPage, callback);
    });
  });
  */
}

function applyPreCheerioFixes(data) {

  var rIncorrectCFText = />ControlFlow</g;
  var fixedCFText = '>Control Flow<';

  var rIncorrectModuleText = />module:(\w+)\.(\w+)</g;

  // the heading needs additional padding at the top so it doesn't get cutoff
  return data
    // for JSDoc to work, the module needs to be labelled 'ControlFlow', while
    // on the page it should appear as 'Control Flow'
    .replace(rIncorrectCFText, fixedCFText)
    // for return types, JSDoc doesn't allow replacing the link text, so it
    // needs to be done here
    .replace(rIncorrectModuleText, function(match, moduleName, methodName) {
      return '>' + methodName + '<';
    });
}


function scrollSpyFix($page, $nav) {
  // move everything into one big ul (for Bootstrap scroll-spy)
  var $ul = $nav.children('ul');
  $ul.addClass('nav').addClass('methods');
  $ul.find('.methods').each(function() {
    var $methodsList = $(this);
    var $methods = $methodsList.find('[data-type="method"]');
    var $parentLi = $methodsList.parent();

    $methodsList.remove();
    $methods.remove();
    $parentLi.after($methods);
    $parentLi.addClass('toc-header');
  });

  $page.find('[data-type="method"]').addClass('toc-method');

  $page.find('[id^="."]').each(function() {
    var $ele = $(this);
    var id = $(this).attr('id');
    $ele.attr('id', id.replace('.', ''));
  });
}

function fixToc(file, $page, moduleFiles) {
  // remove `async` listing from toc
  $page.find('a[href="' + mainModuleFile + '"]').parent().remove();

  // change toc title
  var $nav = $page.find('nav');
  $nav.attr('id', 'toc');
  $nav.children('h3').text(pageTitle);
  $nav.children('h2').remove();

  scrollSpyFix($page, $nav);

  var prependFilename = (file === docFilename) ? '' : docFilename;
  // make everything point to the same 'docs.html' page
  _.each(moduleFiles, function(filename) {
    $page.find('[href^="' + filename + '"]').each(function() {
      var $ele = $(this);
      var href = $ele.attr('href');

      // category titles should sections title, while everything else
      // points to the correct listing
      if (href === filename) {
        var moduleName = $ele.text().toLowerCase().replace(/\s/g, '').replace('.', '');
        $ele.attr('href', prependFilename + '#' + moduleName);
      } else {
        $ele.attr('href', href.replace(filename, prependFilename).replace('#.', '#'));
      }
    });
  });
}

function fixFooter($page) {
  // add a note to the footer that the documentation has been modified
  var $footer = $page.find('footer');
  $footer.append(additionalFooterText);
  $page.find(mainScrollableSection).append($footer);
}

function fixModuleLinks(files, callback) {
  var moduleFiles = extractModuleFiles(files);


  async.each(files, function(file, fileCallback) {
    var filePath = path.join(docsDir, file);
    fs.readFile(filePath, 'utf8', function(err, fileData) {
      if (err) return fileCallback(err);
      var $file = $(applyPreCheerioFixes(fileData));

      var $vDropdown = $file.find('#version-dropdown');
      $vDropdown.find('.dropdown-toggle').contents().get(0).data = 'v' + VERSION + ' ';
      $vDropdown.find('a[href="' + docFilename + '"]').text('v' + VERSION);

      fixToc(file, $file, moduleFiles);
      fixFooter($file);
      $file.find('[href="' + mainModuleFile + '"]').attr('href', docFilename);
      generateHTMLFile(filePath, $file, fileCallback);
    });
  }, callback);
}

//fs.copySync(path.join(__dirname, '../../dist/async.js'), path.join(docsDir, 'scripts/async.js'), { clobber: true });
fs.copySync(path.join(__dirname, './jsdoc-custom.js'), path.join(docsDir, 'scripts/jsdoc-custom.js'),
  { clobber: true });
/*
fs.copySync(path.join(__dirname, '..', '..', 'logo', 'favicon.ico'), path.join(docsDir, 'favicon.ico'), { clobber: true });
fs.copySync(path.join(__dirname, '..', '..', 'logo', 'async-logo.svg'), path.join(docsDir, 'img', 'async-logo.svg'), { clobber: true });
*/

fs.readdir(docsDir, function(err, files) {
  if (err) {
    throw err;
  }

  var HTMLFiles = _.filter(files, function(file) {
    return path.extname(file) === '.html';
  });

  console.log(`html files found in docs: ${HTMLFiles}`);
  async.waterfall([
    function(callback) {
      combineClassDocs(HTMLFiles, function(err) {
        if (err) return callback(err);
        HTMLFiles.push(docFilename);
        return callback(null);
      });
    },
    function(callback) {
      fixModuleLinks(HTMLFiles, callback);
    }
  ], function(err) {
    if (err) throw err;
    console.log('Docs generated successfully');
  });
});
